/*
 * reord.c
 * Copyright (c) 1991 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Dheeraj Sanghi, Department of Computer Science.
 */

/*
 *	reord.c --- Finds out the number of reorders.
 *		It defines reorders as the number of packets that would
 *		need to be stored by a transport level protocol that
 *		guarantees in-sequence delivery. For example, if the packets
 *		are received in the sequence 2,5,3,4, then only 5 is out-of-
 *		sequence, but if the sequence is 2,4,5,3, then both 4 and 5
 *		will need to be stored and are thus counted as reorders.
 *		We ignore all lost packets in calculating reorders.
 *		The input to this program should be packet numbers, one in
 *		each line. There can be other information in the line which
 *		this program ignores.
 */

#include <stdio.h>

#ifdef MIPS
void	*malloc();
void	exit();
#else
char	*malloc();
#endif MIPS

main(argc, argv)
int	argc;
char	*argv[];
{
	int	*reorders;/* reordered packet numbers */
	int	i;		/* index variable */
	int	np;		/* packet number */
	int	nr = 10000;	/* number of records */
	int	rcvnp = 0;	/* no. pf pkts received */
	int	maxnp = 0;	/* highest pkt no. received */
	int	reorder = 0;	/* number of reordered packets */
	int	cumrecv = 0;	/* number of in-sequence packets received */
	char	line[80];	/* store a line of input */
	char	*ptr;		/* poiner to go through options */
	int	*recd;		/* Array of received bits */
	int	*cumloss;	/* Array of cumulative losses */
	int	*seq;		/* Array of sequence number recd. */


	/* option processing */
	argc--; argv++;

	while (argc > 0) {
	    ptr = *argv;
	    while (*ptr) switch (*ptr++) {

	      case '-':
		break;

	      case 'h':		/* help */
	      case 'H':
		usage();

	      case 'n':		/* number of records */
	      case 'N':
		if (*ptr == 0) {
		    argc--; argv++;
		    if (*argv == 0) {
			(void) fprintf (stderr,
"reord: number of records not given with '-n' option\n");
			(void) exit (1);
		    }
		    nr = atoi (*argv);
		}
		else {
		    nr = atoi (ptr);
		    *ptr = 0;
		}
		break;

	      default:
		(void) fprintf (stderr,	"reord: unknown option: '-%c'\n",
				ptr[-1]);
		(void) exit (1);

	    }
	    argc--; argv++;
	}

	/* alloc space for all values */
	recd = (int *) malloc ((unsigned) ((sizeof (int))*(nr+10)));
	seq = (int *) malloc ((unsigned) ((sizeof (int))*(nr+10)));
	cumloss = (int *) malloc ((unsigned) ((sizeof (int))*(nr+10)));
	reorders = (int *) malloc ((unsigned) ((sizeof (int))*(nr+10)));

	/* initialize variables */
	for (i = 0; i <= nr; i++) {
	    recd[i] = 0;
	    seq[i] = 0;
	    cumloss[i] = 0;
	}

	/* Get a line of input which has a pkt no. as the first field */
	while (gets(line)) {
	    (void) sscanf (line, "%d", &np);
	    if (maxnp < np)
		maxnp = np;		/* save highest pkt-id */

	    /* maxm pkt number should be less than no of records */
	    if (maxnp > nr) {
		(void) fprintf (stderr, "number of records > %d, ", nr);
		(void) fprintf (stderr, "ignoring other records\n");
		(void) fprintf (stderr,
			"you may rerun with higher number of records\n");
		break;
	    }

	    /* if duplicate then ignore */
	    if (recd[np]) continue;

	    rcvnp++;			/* one more pkt received */
	    recd[np] = 1;		/* pkt 'np' is received */
	    seq[rcvnp] = np;		/* save pkt-ids in order of arrival */
	}

	/* calculate cumulative losses until each pkt-id */
	for (i=1; i <= maxnp; i++) {
		if (recd[i])
			cumloss[i] = cumloss[i-1];
		else
			cumloss[i] = cumloss[i-1] + 1;
	}

	/*
	 * now we will ignore the lost packets. for this, we subtract
	 * from pkt-ids the number of cumulative losses until that pkt.
	 * the result will be a sequence of pkt-ids in the order of
	 * their arrival in which there are no lost packets.
	 * we reset the 'recd' bit since that needs to calculated based
	 * on new effective packet-ids. we will do that next.
	 */
	for (i = 1; i <= maxnp; i++) {
		seq[i] = seq[i] - cumloss[seq[i]];
		recd[i] = 0;
	}

	/* for all packets received */
	for (i = 1; i <= rcvnp; i++) {

		/* set the received bit for packets in order of arrival */
		recd[seq[i]] = 1;

		/* if the current sequence number is greater than
		 * the highest consecutive sequence number seen, than
		 * this packet is clearly a reorder.
		 */
		if (seq[i] > cumrecv + 1)
			{
			reorder++;
			reorders[reorder] = seq[i];
			}
		else
			while (recd[cumrecv + 1]) cumrecv++;
	}

	(void) fprintf (stdout, "Number of reorders = %d\n", reorder);

	/*changed */
	(void)fprintf(stdout, "Reordered packets are : "); 
	(void)fprintf(stdout, "\n");
	for(i=1;i<=reorder;i++)
	(void)fprintf(stdout, "%d ", reorders[i]);
	(void)fprintf(stdout, "\n");
	/* changed */
}

usage()
{
    (void) fprintf (stderr, "usage: reord -n #packets\n");
    (void) exit (1);
}
