/*
 * udprecv.c
 * Copyright (c) 1991 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Dheeraj Sanghi, Department of Computer Science.
 */

/*
 *	udprecv.c --- The receiver program. It receives packets from either
 *		the sender program or the udpecho program. It adds its own
 *		timestamp to the packet, and sends the information added
 *		by sender and udpecho to the logger process, udpread.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>

#include "udpexpt.h"

/*### The following definition was not made on BSDish systems */
#ifndef SOLARIS
typedef unsigned int    in_addr_t;
#endif

in_addr_t gethostaddr();

#ifdef MIPS
void	*malloc();
void	exit ();
void	bzero ();
#else
char	*malloc();
#endif

#define MAXHNAME      100             /* maximum hostname size */
#define PKTSIZE		32		/* default packet size */
#define INDICATOR_SIZE	32		/* Size of the indicator to */
					/* signal temination of accept */

main(argc, argv)
     int	argc;
     char	*argv[];
{
   struct	sockaddr_in src;	/* socket addr for incoming packets */
   struct	sockaddr_in dst;	/* socket addr for outgoing packets */
   int	soudp;			/* socket number for incoming conn. */
   int	sotcp;			/* socket number for outgoing conn. */
   /*
    * We could have used sockaddr data structure and that would have
    * allowed us with minor modifications, the ability to use other
    * domains like UNIX, NS etc. but since we are only interested in
    * doing expt over the Internet, we don't care for such flexibility. 
    * Here the incoming connection is on UDP and the outgoing connection
    * is on TCP since we wanted reliability in storing the results.
    */
   char	data[MAXMESG];		/* Maxm size packet */
   char	*ptr;			/* to traverse the arguments */
   int	rcv_size = MAXRCVBUF;	/* socket buffer size for receive */
   int	np;			/* packet number received */
   struct	timeval *tp;		/* timestamp */
   int	twoway = YES;		/* is conn. 1-way or 2-way ? */
   struct	udprecord *udprecord;	/* log record for a packet */
   int	*fromlen;
   /* AB Ozgur */
   char    hostname[MAXHNAME];     /* udpread hostname */
   int     readhost = NO;          /* is udpread hostname given ? */
   /* EoAB Ozgur */

   /* Addind the from field for the TCP source connection. Use in */
   /* the accept system call. -- Suman */
   struct sockaddr from;
   int new_sotcp;
   int verbose = NO;
   int pktsize = PKTSIZE;
   char indicator[INDICATOR_SIZE];
   int ds_ptr = 0;
   char *data_store;
/*
   int fragmented = NO;
*/

   /* alloc space for timestamp */
   tp = (struct timeval *)(malloc (sizeof (struct timeval)));

   fromlen = (int *) malloc (sizeof (int *));

   argc--; argv++;

   /* go through the arguments */
   while (argc > 0) {
      ptr = *argv;
      while (*ptr) switch (*ptr++) {

      case '-':
	 break;

      case '1':
	 twoway = NO;	/* one-way connection, most probably
			 * using IP option to route the packet
			 * No echo process. */
	 break;

      case '2':
	 twoway = YES;	/* two-way connection. There is an
			 * echo process. This is default. */
	 break;

      case 'l':            /* Logger host where udpread is running */
      case 'L':
	 readhost = YES;   /* final host is given */
	 if (*ptr == 0) {
	    argc--; argv++;
	    if (*argv == 0) {
	       (void) fprintf (stderr,
			       "udprecv: no host for udpread is given with '-l|L'.\n");
	       (void) exit (1);
	    }
	    (void) strcpy (hostname, *argv);
	 }
	 else {
	    (void) strcpy (hostname, ptr);
	    *ptr = 0;
	 }
	 break;

      case 'v': 	/* verbose */
      case 'V':
	 verbose = YES;
	 break;

      case 'p':		/* packet size */
      case 'P':
      	if (*ptr == 0) {
        argc--; argv++;
        if (*argv == 0) {
         	(void) fprintf (stderr,
                        "tcpsend: no packet size given with '-p'.\n");
                (void) exit (1);
                }
                pktsize = atoi (*argv);
        }
        else {
        	pktsize = atoi (*argv);
                *ptr = 0;
        }
	if (MAXMESG < pktsize)
		usage();
        break;

      default:
	 (void) fprintf (stderr,
			 "udprecv: Unknown option '%c'\n", ptr[-1]);

      case 'h':
      case 'H':
	 usage();
      }
      argc--; argv++;
   }

   /* The first part of the data packet can be accessed
    * as the log record */
   udprecord = (struct udprecord *) data;

   /* create a socket to receive UDP packets */
   soudp = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
   if (soudp < 0) {
      perror("tcprecv: socket");
      exit(1);
   }

   /* create a socket to send result packets by TCP */
   sotcp = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
   if (sotcp < 0) {
      perror("udprecv: tcp socket");
      exit(1);
   }

   /* initialize the socket address for outgoing connection */
   bzero ((char *)&dst, sizeof(dst));
   dst.sin_family = AF_INET;
   dst.sin_port = htons(READPORT);
   /*
    * loopback interface is 127.0.0.1
    * We are assuming that the reader process will always run
    * on the same host. If we want the flexibility to have the
    * reader run on a different host, then we need to get
    * destination as one of the option. Then we need to set
    * the IP address of the destination as sin_addr here.
    =>
    Mentioned flexibility AB Ozgur
    =>
    */
   if (readhost == NO)     /* No udpread host defined*/
      dst.sin_addr.s_addr = htonl((u_long) 0x7F000001);
/* The above line was causing an "Unaligned access" error if it were
   in the following original format:
 *(long *)&dst.sin_addr = htonl((u_long) 0x7F000001);
*/
   else
      if ( (dst.sin_addr.s_addr = gethostaddr(hostname)) == 0 ) {
	 (void) fprintf (stderr, "udprecv: %s: unknown host\n",
			 hostname);
	 (void) exit (1);
      }

   /* connect to the reader process */
   if (connect(sotcp, (struct sockaddr *)&dst, sizeof(dst)) < 0) {
      perror("udprecv: connect");
      exit (1);
   }

   /* initialize socket address for incoming connection */
   bzero((char *)&src, sizeof (src));
   src.sin_family = AF_INET;
   src.sin_port = htons(RECVPORT);

   /* bind the address (port#, family) to the socket */
   if (bind (soudp, (struct sockaddr *) &src, sizeof(src)) < 0) {
      perror ("udprecv: bind");
      exit (1);
   }

   /* For TCP, adding the listen and accept calls -- Suman. */
   if (listen(soudp,1) < 0) {
	perror("tcprecv: listen");
 	exit(1);
   }

   *fromlen = (int)sizeof(struct sockaddr);

   if ((new_sotcp = accept(soudp,&from,fromlen)) < 0) {
	perror("tcprecv: accept");
	exit(1);
   }

   /* set the socket receive buffer to maximum possible.
    * this will minimize the chance of losses at the endpoint */
   if (setsockopt(new_sotcp, SOL_SOCKET, SO_RCVBUF, (char *)&rcv_size,
		  sizeof (rcv_size)) < 0) {
      perror ("udprecv: receive buffer option");
      exit (1);
   }

/* Send the indicator from the src/echo host indicating that */
/* it is ready to receive the messages now. -- Suman */
   if (write (new_sotcp,indicator,INDICATOR_SIZE) < 0)
	{
	 perror ("tcprecv: write");
	 exit(1);
	}


  ds_ptr = 0;
  data_store = (char *) malloc (sizeof(pktsize));

   while (1) {
      static 	int incc, outcc;		/* char count for recv/send */
	int data_pos;
	int to_break;
	static int udprec_size = (int)sizeof(struct udprecord);
	int new_incc;

        /* Before we read in a packet from the sender, we put the */
        /* stuff from data_store in there, so that any partial */
        /* packet that we received till now, will be contiguous */
        if (ds_ptr > 0)
                 memcpy(data,data_store,ds_ptr);


      /* read one packet */
      new_incc = read (new_sotcp, (char*)&(data[ds_ptr]), MAXMESG-ds_ptr);
      if (incc < 0) {
	 perror ("udprecv: read");
	 break;
      }

      incc = new_incc + ds_ptr;

/* Some debug code --  Suman */
	if (verbose)
		fprintf(stdout,"Recvd %d bytes\n",incc);

      ds_ptr = 0;

      /* get the timestamp */
      (void) gettimeofday (tp, (struct timezone *) 0);

	data_pos = 0;
	to_break = NO;

	while ((incc - data_pos) >= pktsize)
		{
		 udprecord = (struct udprecord *)&(data[data_pos]);

      		/* add this timestamp to the log record of this packet */
      		 if (twoway)
			{
	 	 	udprecord->rcv.tv_sec = htonl ((u_long) tp->tv_sec);
	 		udprecord->rcv.tv_usec = htonl ((u_long) tp->tv_usec);
      		 	}
      	 	 else 			
			{
	 		udprecord->midhost.tv_sec = htonl ((u_long) tp->tv_sec);
	 		udprecord->midhost.tv_usec = htonl ((u_long) tp->tv_usec);
      		 	}
	/* Debug code */
		{
		 static int last_np;
		 last_np = np;

/*@@@@@ Dont delete next line */
      		/* get the packet number */
      		 np = (int) ntohl ((u_long) udprecord->num);
/*@@@@@ Dont delete prev line */

/*
		 if ((np < 0) || (np > 45000))
			printf ("Have a problem --- %d is pkt num Prev was %d\n",np,last_np);
*/
		}

		/* What happens if only part of the write goes thru ? */
		/* have to repeat the write for the rest of the packet */
      		/* send just the log record to the reader */

		 outcc = 0;
		 do
			{
			 int new_outcc;

      		 	 new_outcc = write (sotcp,
					 (char *)&(((char*)udprecord)[outcc]),
		  				udprec_size - outcc);
      		 	 if (new_outcc < 0)
				{
	 			 perror("udprecv: tcp write");
				 to_break = YES;
	 			 break;
      				}

			 outcc += new_outcc;
			}
		 while (outcc < udprec_size);

		 if (to_break)
			break;

	/* some more debug code */
		 if (outcc < (int)sizeof (struct udprecord))
			{
			 fprintf (stderr,"ALERT ! WROTE OUT LESS THAN A UDPRECORD on pkt %d  !\n",np);
			}

      /*
       * to end the connection, sender program will be sending a
       * few packets in the end with packet number 0. It sends
       * more than 1, so that at least one of them makes it to
       * the receiver and the receiver process can stop gracefully.
       */
      		 if (!np)
			{
	 		(void) fprintf (stderr,
			 	"udprecv: Finished servicing this client\n");
			to_break = YES;
	 		break;
      			}

		 data_pos += pktsize;
		}

	if (to_break)
		break;

	if (data_pos != incc)
		{
                 memcpy(data_store,(char*)&(data[data_pos]),incc-data_pos);
                 ds_ptr = incc - data_pos;
		 fprintf (stderr,"Pkt %d fragmented and recvd %d bytes\n",(np+1),new_incc);
                }
   }
   free(data_store);
   /* close the open sockets */
   (void) close (sotcp);
   (void) close (new_sotcp);
   (void) close (soudp);

}


usage()
{
   (void) fprintf (stderr,"usage: tcprecv -[1|2] -l <logger host> [-p <packet size>] [-v]\n");
   (void) fprintf (stderr,"Keep MAXMESG (%d) >= packet size \n",MAXMESG);
   exit (1);
}

/*
  CopiedB Ozgur from udpsend.c
*/
 
/*
 * Return the address (in network byte order) of the requested host.
 *
 * We return 0 if we can't find anything.
 *
 * Arguably, this should return a struct in_addr, but 4.3 doesn't
 * seem to be able to make up it's mind.
 */

in_addr_t
gethostaddr(name)
     char *name;
{
   in_addr_t addr;
   register struct	hostent *hp;

   if ((addr = (in_addr_t)inet_addr (name)) != -1)
      return (addr);

   hp = gethostbyname(name);
   if (hp != NULL)
      return (*(in_addr_t *)hp->h_addr);
   else
      return (0);
}
