/*
 * udpecho.c
 * Copyright (c) 1991 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Dheeraj Sanghi, Department of Computer Science.
 */

/*
 *	udpecho.c --- This program runs on the remote host. This receives
 *		packets from the sender program and sends packets to the
 *		receiver program. It adds its own timestamp and the number
 *		of packets it has recived so far, to the packet.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "udpexpt.h"

#ifdef MIPS
void	*malloc();
void	exit();
void	bzero();
#else
char	*malloc();
#endif MIPS

main(argc, argv)
int	argc;
char	*argv[];
{
	struct	sockaddr_in src;	/* Own address/port etc. */
	struct	sockaddr_in dst;	/* Reciver's address/port */
	struct	sockaddr_in fromaddr;	/* sender's address/port */
	int	fromlen;		/* size of sender's address/port */
	int	so1;			/* socket id for incoming pkts */
	int	so2;			/* socket id for outgoing pkts */
	char	data[MAXMESG];		/* maximum size packet */
	int	rcv_size = MAXRCVBUF;	/* socket receive buffer size */
	int	np;			/* sender's packet number */
	int	necho = 0;		/* number of packets recd */
	struct	timeval *tp;		/* local timestamp */
	struct	udprecord *udprecord;	/* log record for the packet */
	char	*ptr;			/* to traverse the arguments */

	/* alloc space for local timestamp */
	tp = (struct timeval *)(malloc (sizeof (struct timeval)));

	argc--; argv++;

	/* go through the arguments */
	while (argc > 0) {
		ptr = *argv;
		while (*ptr) switch (*ptr++) {

		case '-':
			break;

		default:
			(void) fprintf (stderr,
		"udpecho: Unknown option '%c'\n", ptr[-1]);

		case 'h':		/* help */
		case 'H':
			usage();
		}

		argc--; argv++;
	}

	/* access initial part of packet as log record */
	udprecord = (struct udprecord *) data;

	/* size of socket address data structure for from address */
	fromlen = sizeof (fromaddr);

	/* create a socket for conenction with sender */
	so1 = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (so1 < 0) {
		perror("udpecho: socket");
		exit(1);
	}

	/* create a socket for connection with receiver */
	so2 = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (so2 < 0) {
		perror("udpecho: socket");
		exit(1);
	}

	/* initialize address/port etc for incoming connection */
	bzero((char *)&src, sizeof (src));
	src.sin_family = AF_INET;
	src.sin_port = htons(MIDPORT);

	/* bind the address/port number to the socket */
	if (bind (so1, (struct sockaddr *) &src, sizeof(src)) < 0) {
		perror ("udpecho: bind");
		exit (1);
	}

	/* initialize address/port for outgoing connection */
	bzero((char *)&dst, sizeof (dst));
	dst.sin_family = AF_INET;
	dst.sin_port = htons(RECVPORT);

	/* set receive buffers to maximum so that packets are not lost if
	 * for some reason this program does not get CPU for some time. */
	if (setsockopt(so1, SOL_SOCKET, SO_RCVBUF, (char *)&rcv_size,
			sizeof (rcv_size)) < 0) {
		perror ("udpecho: receive buffer option");
		exit (1);
	}

	/* repeat indefinitely */
	/* TO DO: There should be some clean mechanism to kill the echo
	 * process from remote. Well, may be not. May me it should run as
	 * daemon all the time. Anyway, right now, it can be killed only
	 * by sending a kill signal.
	 */
	while (1) {
		static 	int incc, outcc;

		/* receive the packet from the sender */
		incc = recvfrom (so1, data, MAXMESG, 0,
				(struct sockaddr *)&fromaddr, &fromlen);
		if (incc < 0) {
			perror ("udpecho: read");
			break;
		}
		/* set the destination address from the 'from' address */
		dst.sin_addr = fromaddr.sin_addr;

		/* Increment the number of packets received */
		necho++;

		/* get the local timestamp */
		(void) gettimeofday (tp, (struct timezone *) 0);

		/* Add local timestamp and echo number to log record */
		udprecord->midhost.tv_sec = htonl ((u_long) tp->tv_sec);
		udprecord->midhost.tv_usec = htonl ((u_long) tp->tv_usec);
		udprecord->echonum = htonl ((u_long) necho);

/* Debugging for Kernel timing stuff -- SB. */
/*
		printf ("Packet %d at kernel :%u . %u\n",udprecord->num,udprecord->snd.tv_sec,udprecord->snd.tv_usec);
		printf ("Packet %d at user :%u . %u\n",udprecord->num,udprecord->midhost.tv_sec,udprecord->midhost.tv_usec);
*/

		/* get the sender's packet number */
		np = ntohl ((u_long) udprecord->num);

		/* send it out to the receiver */
		outcc = sendto (so2, data, incc, 0,
					(struct sockaddr *)&dst, sizeof (dst));
		if (outcc < 0)
			perror ("UDPecho: sendto");
		/*
		 * If the incoming packet has packet-id of 0, then it is to
		 * tell the receiver to stop expecting more packets.
		 * Reset echo number to 0, since the next packets would
		 * be on a new experiment.
		 */
		if (!np) necho = 0;
	}
}


usage()
{
	(void) fprintf (stderr, "usage: udpecho\n");
	(void) exit (1);
}

