/*
 * udpread.c
 * Copyright (c) 1991 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Dheeraj Sanghi, Department of Computer Science.
 */

/*
 *	udpread.c --- The reader program. Udprecv sends all the
 *		information regarding packets that are received by it, to
 *		this program. Udpread program writes that information
 *		on a disk file in ASCII format, which can be easily
 *		processed by other tools, such as awk.
 */

#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "udpexpt.h"

#ifdef MIPS
void	exit ();
void	bzero ();
#endif

#define MAXPACKETS	300010		/* max no. of pkts in a connection */
struct packetsStruct {
   int num;			/* packet numbers */
   int echoNum;			/*  echo number to understand loss path */
   double rtt;			/* roundtrip time */
} packets[MAXPACKETS];

int
main(argc, argv)
int	argc;
char	*argv[];
{
   struct	sockaddr_in src;	/* our address/port */
   struct	sockaddr_in dst;	/* remote address/port */
   int	nso, so;		/* socket ids */
   char	*ptr;			/* pointer to go through arguments */
   int	rcv_size = MAXRCVBUF;	/* socket receive buffer size */
   struct	stat	statb;		/* info on output file */
   int	oflag = NO;		/* whether file name given ? */
   FILE	*fd;			/* output file descriptor */
   char	outfile[MAXFNAME];	/* output file name */
   int	rsize;			/* size of sockaddr_in */
   int	initsecsSnd, initsecsMid;     	/* 1st timestamp - to normalize */
   double	snd1, mid1, rcv1;	/* 3 timestamps in log record */
   struct	udprecord udprecord;	/* log record */
   /* Followin vars are xferred from udprecv ### B Ozgur */
   int	maxnp = 0;		/* highest packet number received */   
   int	total = 0;		/* total number of packets received */

/* Debug variables - last_np and np */
 static int last_np;
 static int np;

   /* Followin vars are xferred from udprtt ### B Ozgur */
   /* As the rtt calculation is shifted back, we may not need an
      implementation with arrays but to count for lost or out-of-order
      packets, they are kept.   
      (Originally num[] and rtt[] were defined here too (instead of
      being global) but, somehow when the debugger option of cc is
      turned on we get 'Segmentation fault even before starting the
      program(main)
   */
   int	i;		        	/* index variable */
   /* New variables to find duplicates and losses */
   int j, nDuplicates=0, nSendLosses=0, nRetLosses=0, lastEchoNum;
   
   (void) bzero (packets, sizeof(struct packetsStruct) * MAXPACKETS);

   /* argument processing */
   argc--; argv++;

   while (argc > 0) {
      ptr = *argv;
      while (*ptr) switch (*ptr++) {

      case '-':
	 break;

      case 'o':
      case 'O':
	 oflag = YES;
	 if (*ptr == 0) {
	    argc--; argv++;
	    if (*argv == 0) {
	       (void) fprintf (stderr,
			       "udpread: no output file given with '-o'.\n");
	       (void) exit (1);
	    }
	    (void) strcpy (outfile, *argv);
	 }
	 else {
	    (void) strcpy (outfile, ptr);
	    *ptr = 0;
	 }
	 break;

      default:
	 (void) fprintf (stderr,
			 "udpread: Unknown option '%c'\n", ptr[-1]);

      case 'h':
      case 'H':
	 usage();

      }
      argc--; argv++;
   }

   /* If no output file given, ask for it */
   if (!oflag) {
      (void) fprintf (stdout, "Output File? ");
      (void) fflush (stdout);
      (void) gets(outfile);
   }

   /* filename "-" is same as standard output */
   if (!(strcmp (outfile, "-"))) {
      fd = stdout;
      goto nooutfile;
   }

   if (stat (outfile, &statb) < 0)
      /* If the file does not exist */
      fd = fopen(outfile, "w");
   else
      /* File exists; may be check for directory too */
      fd = fopen(outfile, "a");
   /*
    * To Do: At this time, if the file exists, instead of just
    * opening it in the append mode, may be we should ask the
    * user if he wants to overwrite, or give another file name etc.
    */

   if (fd == NULL) {
      (void) fprintf (stderr, "can't open file %s\n", outfile);
      perror("send: fopen");
      (void) exit (1);
   }

nooutfile:

   /* create a socket for receiving log records */
   so = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
   if (so < 0) {
      perror("udpread: tcp socket");
      (void) exit (1);
   }

   /* initialize address/port for that socket */
   (void) bzero ((char *)&src, sizeof (src));
   src.sin_family = AF_INET;
   src.sin_port = htons((u_short) READPORT);

   /* bind address/port to socket */
   if (bind (so, (struct sockaddr *) &src, sizeof(src)) < 0) {
      perror ("udpread: bind");
      (void) exit (1);
   }

   (void) listen (so, 2);
   rsize = sizeof(dst);

   /* accept the connection */
   nso = accept (so, (struct sockaddr *)&dst, &rsize);
   if (nso < 0) {
      perror ("udpread: accept");
      (void) exit (1);
   }

   /* set socket receive buffer size to maximum so that we don't
    * get stuck because of lack of memory */
   if (setsockopt (nso, SOL_SOCKET, SO_RCVBUF, (char *) &rcv_size,
		   sizeof (rcv_size)) < 0) {
      perror ("udpread: receive buffer option");
      (void) exit (1);
   }

   /* we can close the original socket now */
   (void) close (so);

   /* repeat until connection is broken by the receiver */
   while (1) {
      static int cc;			/* character count */
      static int init = YES;		/* is it first packet ? */
      static int udprec_size = (int)sizeof(struct udprecord);

  /* Possible bug fix -- what happens if the data read is less than */
  /* a udprecord. Would have to loop and get the rest. -- Suman */
      /* read one log record */
     cc = 0;
     do
	{
	 int new_cc;

      	 new_cc = read (nso, (char *) &(((char *)&udprecord)[cc]),
		 udprec_size - cc);
      	 if (new_cc < 0) {
	 	perror ("udpread: read");
	 	break;
      		}

	 cc += new_cc;
	}
     while (cc < udprec_size);

/* some debug code */
	if (cc < (int)sizeof (struct udprecord))
		{
		 fprintf (stderr,"ALERT !!!!!!! READ LESS ... PREV WAS %d\n",last_np);
		}

      /* if the connection is closed by the peer */
      if (cc == 0) {
	 (void) fprintf (stderr,"udpread: read returned 0\n");
	 break;
      }

#define NTOH(a) a = ntohl ((u_long) (a))
      NTOH(udprecord.snd.tv_sec);
      NTOH(udprecord.snd.tv_usec);
      NTOH(udprecord.midhost.tv_sec);
      NTOH(udprecord.midhost.tv_usec);
      NTOH(udprecord.rcv.tv_sec);
      NTOH(udprecord.rcv.tv_usec);
      NTOH(udprecord.num);
      NTOH(udprecord.echonum);
#undef NTOH

/* debug code */
 {
 last_np = (int) np;
 np = (int) udprecord.num;
/*
 if ((np < 0) || (np > 45000))
	(void) fprintf (stderr,"Have a problem %d %d\n",np,last_np);
*/
 }

      /* if this is the first packet */
      if (init) {
	 initsecsSnd = udprecord.snd.tv_sec	- 1000;
	 initsecsMid = udprecord.midhost.tv_sec	- 1000;
	 /* we normalize time so that the send-time for the first
	  * packet is 1000 second. we could have started with 0, but
	  * if there is a reordering we could end up with negative
	  * time later. The post-processing programs will assume all
	  * times to be positive.
	  */
	 lastEchoNum = udprecord.echonum;
	 init = NO;
      }

      /* record number 0 is indicative of finished connection */
      if (udprecord.num == 0) break;

      /* Now we have a normal packet at hand. Do the
         calculations/checks to be written in summary file */
      /* set the maximum packet number received */
      if (maxnp < udprecord.num)
	 maxnp = udprecord.num;
      total++;		/* one more packet received */

      /* normalize the three timestamps. this will help in
       * reducing the space requirement to store data */
      snd1 = (udprecord.snd.tv_sec - initsecsSnd) +
	 (udprecord.snd.tv_usec / 1000000.0);
      mid1 = (udprecord.midhost.tv_sec - initsecsMid) +
	 (udprecord.midhost.tv_usec / 1000000.0);

      /* only if it was 2-way connection */
      if (udprecord.rcv.tv_sec)
	 rcv1 = (udprecord.rcv.tv_sec - initsecsSnd) +
	    (udprecord.rcv.tv_usec / 1000000.0);
      else
	 rcv1 = 0.0;
      /* Now set the flag for this packet and find its RTT */
      if (packets[udprecord.num].num == 0) {
	 packets[udprecord.num].num = udprecord.num;
	 packets[udprecord.num].echoNum = udprecord.echonum;
	 packets[udprecord.num].rtt = ( (udprecord.rcv.tv_sec) ? rcv1 : mid1) - snd1;
      }
      else
	 nDuplicates++;

      /* output the log record */
      (void) fprintf (fd, "%d %.6lf %.6lf %.6lf %d\n",
		      (int) udprecord.num, snd1, mid1, rcv1,
		      (int) udprecord.echonum);
   }
   (void) close (nso);
   (void) fclose (fd);

   /* Now use the same fd to write out rtt file */
   if ( (fd=fopen(strcat(outfile, ".rtt"), "w")) == NULL) {
      (void) fprintf (stderr, "can't open file %s\n", outfile);
      perror("send: fopen");
      (void) exit (1);
   }
   for (i = 1; i <= maxnp; i++)
      /* print the roundtrip times */
      if (packets[i].num > 0) {
	 (void) fprintf (fd, "%d %9.6lf\n", i, packets[i].rtt);
	 lastEchoNum = packets[i].echoNum;
      }
      else {	/* roundtrip time is reported as -0.001 for lost packets */
	 (void) fprintf (fd, "%d %6.3lf\n", i, -0.001);
	 for (j=i+1; packets[j].echoNum <= 0; j++); /* Guanteed to terminate */
	 if (packets[j].echoNum - lastEchoNum == 1)
	    nSendLosses++;
	 else {
	    nRetLosses++;
	    lastEchoNum++;
	 }
      }
   
   /* print a bit of stats to stdout -> Summary file */
   (void) fprintf (stdout, "Highest packet no. received = %d\n", maxnp);
   (void) fprintf (stdout, "Number of packets received = %d\n", total);
   (void) fprintf (stdout, "Number of duplicate packets = %d\n", nDuplicates);
   (void) fprintf (stdout, "Number of losses = %d\n", nSendLosses+nRetLosses);
   (void) fprintf (stdout, "Number of losses on send path = %d\n", nSendLosses);
   (void) fprintf (stdout, "Number of losses on receive path = %d\n", nRetLosses);
   (void) fclose (fd);
   
   return 0;
}

usage()
{
   (void) fprintf (stderr,"usage: udpread [-h] [-o outfile]\n");
   (void) exit (1);
}

