/*
 * udpsend.c
 * Copyright (c) 1991 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Dheeraj Sanghi, Department of Computer Science.
 */

/*
 *	udpsend.c --- The sender program, sends a specified number of
 *		packets to either the udpecho or the udprecv process
 *		running elsewhere. The interpacket delay is specified by
 *		the user. The packets can be routed via a remote gateway
 *		using IP_LSRR option, if so desired, but its use is
 *		discouraged. Size of the user packet can be specified.
 *		Information about the experiment is saved in a file.
 */

#include <strings.h>
#include <stdio.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/param.h>
#include <sys/time.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>

#include "udpexpt.h"

#define PKTSIZE		32		/* default packet size */
#define NPKT		1000		/* default number of packets */
#define MAXHNAME	100		/* maximum hostname size */

/*
 * structure 'rr' is used to store the IP_LSRR option, if that is the
 * mode of operation. Use of this option is not recommended since many
 * intermediate gateways give lower priority to the IP-optioned packets.
 * This software should be used, as far as possible, with an actual
 * process (receive or echo) running on the remote host.
 */

static struct rr_opt {
   u_char      rr_nop1;		/* 0 */
   u_char      rr_lsrr;		/* 1 */
   u_char      rr_lsrr_len;		/* 2 */
   u_char      rr_lsrr_ptr;		/* 3 */
   struct in_addr  rr_lsrr_gwy;	/* 4567 */
   struct in_addr  rr_lsrr_dst;	/* 89ab */
} rr;

/*### The following definition was not made on BSDish systems */
#ifndef SOLARIS
typedef unsigned int    in_addr_t;
#endif

in_addr_t gethostaddr();

#ifdef MIPS
void	*malloc();
void	bzero();
void	exit();
unsigned	sleep();
#else
char	*malloc();
#endif

main(argc, argv)
     int	argc;
     char	*argv[];
{
   struct	sockaddr_in dst;	/* destination internet address */
   struct	timeval	*tp;		/* used for getting timeofday */
   struct	stat	statb;		/* to see if output file exists */
   int	so;			/* socket to remote host */
   FILE	*fd;			/* info file descriptor*/
   char	data[MAXMESG];		/* maximum size data packet */
   int	verbose = NO;		/* inform about progress */
   int	gway = NO;		/* is there a gateway ? default NO */
   int	dsthost = NO;		/* is final destination given ? */
   int	midhost	= NO;		/* is echo destination given  ? */
   /*
    * only one of the echo destination or final destination can be
    * given. if none is given it is assumed that final destination
    * is the same host. gateway host is used to specify the route.
    * IP options are used to route the packet to the
    * echo/final destination via a given gateway.
    */
   int	oflag = NO;		/* is info file name given ? */
   int	sleepflag = NO;		/* is there sleep between packets */
   int	pktsize = PKTSIZE;	/* packet size */
   int	npkt = NPKT;		/* number of packets */
   int	nbulk = 1;		/* number of packets sent-together */
   int	ntrain = 1;		/* number of packet trains */
   /*
    * the packets are sent as follows: we send 'ntrain' number of
    * packet trains. In each train, we send packets 'npkt' number
    * of times. Each time, we send 'nbulk' number of packets.
    * thus a total of nbulk * npkt * ntrain packets get sent.
    */
   int	nadd = 0;		/* dummy additions between 2 bulks */
   int	gap = 0;		/* gap in seconds between 2 trains */
   unsigned	sleeptime = 0;	/* sleep in ms between 2 bulks */
   /*
    * the way send time is controlled is as follows: a bulk is sent as
    * fast as possible. between two successive bulks of the same train,
    * we do 'nadd' number of dummy additions. this provides finer
    * control on inter-packet send-time delay, but causes cpu wastage,
    * and may result in this process getting lower priority later on.
    * we then go for sleep for 'sleeptime' ms. Between two successive
    * trains, we go for a 'gap' amount of sleep (in seconds).
    */
   int	pktsend = 0;		/* number of packets sent so far */
   int	write_error = NO;	/* error on sending a packet */
   char	hostname[MAXHNAME];	/* echo/final destination hostname */
   char	gateway[MAXHNAME];	/* name of the gateway host */
   char	outfile[MAXFNAME];	/* output (info) file name */
   int	dcnt = 0;		/* count of packets for verbose */
   int	dcmax = 0;		/* output diagnosis every dcmax pkts */
   int	nc;			/* index variable for packet-count */
   int	nt;			/* index variable for train-count */
   int	nb;			/* index variable for bulk-count */
   char	*ptr;			/* pointer to traverse options */
   struct	udprecord *udprecord;	/* log record */

   /* allocate space for local timestamp */
   tp = (struct timeval *)(malloc (sizeof (struct timeval)));

   argv++; argc--;

   /* initialize the IP option record.
    * we will use the whole thing only with gateway option */
   rr.rr_nop1 = IPOPT_NOP;
   rr.rr_lsrr = IPOPT_LSRR;
   rr.rr_lsrr_len = 2 * sizeof (struct in_addr) + 3;
   rr.rr_lsrr_ptr = IPOPT_MINOFF;
   rr.rr_lsrr_gwy.s_addr = 0;
   rr.rr_lsrr_dst.s_addr = 0;

   /* option processing */
   while (argc > 0) {
      ptr = *argv;
      while (*ptr) switch (*ptr++) {

      case '-':
	 break;

      case 'b':		/* bulk size */
      case 'B':
	 if (*ptr == 0) {
	    argc--; argv++;
	    if (*argv == 0) {
	       (void) fprintf (stderr,
			       "udpsend: no bulk size given with '-b'.\n");
	       (void) exit (1);
	    }
	    nbulk = atoi(*argv);
	 }
	 else {
	    nbulk = atoi (ptr);
	    *ptr = 0;
	 }
	 break;

      case 'c':		/* packet count */
      case 'C':
	 if (*ptr == 0) {
	    argc--; argv++;
	    if (*argv == 0) {
	       (void) fprintf (stderr,
			       "udpsend: no packet count given with '-c'.\n");
	       (void) exit (1);
	    }
	    npkt = atoi(*argv);
	 }
	 else {
	    npkt = atoi (ptr);
	    *ptr = 0;
	 }
	 break;

      case 'm':		/* echo host is given */
      case 'M':
	 midhost = YES;

      case 'd':
      case 'D':
	 dsthost = YES;	/* final host is given */
	 if (*ptr == 0) {
	    argc--; argv++;
	    if (*argv == 0) {
	       (void) fprintf (stderr,
			       "udpsend: no destination host given with '-d|m'.\n");
	       (void) exit (1);
	    }
	    (void) strcpy (hostname, *argv);
	 }
	 else {
	    (void) strcpy (hostname, ptr);
	    *ptr = 0;
	 }
	 break;
	 /* the way echo and final destination options
	  * have been implemented, echo host option
	  * supercedes the final destination option.
	  * the program would not complain if you try to
	  * give both options, but would silently ignore
	  * the final destination option.
	  */

      case 'g':		/* gateway */
      case 'G':
	 gway = YES;
	 if (*ptr == 0) {
	    argc--; argv++;
	    if (*argv == 0) {
	       (void) fprintf (stderr,
			       "udpsend: no gateway given with '-g'.\n");
	       (void) exit (1);
	    }
	    (void) strcpy (gateway, *argv);
	 }
	 else {
	    (void) strcpy (gateway, ptr);
	    *ptr = 0;
	 }
	 break;

      case 'h':		/* help */
      case 'H':
	 usage();

      case 'i':		/* inter-train gap in seconds */
      case 'I':
	 if (*ptr == 0) {
	    argc--; argv++;
	    if (*argv == 0) {
	       (void) fprintf (stderr,
			       "udpsend: inter-train gap not given with '-i'.\n");
	       (void) exit (1);
	    }
	    gap = atoi(*argv);
	 }
	 else {
	    gap = atoi (ptr);
	    *ptr = 0;
	 }
	 break;

      case 'n':		/* number of dummy additions */
      case 'N':
	 if (*ptr == 0) {
	    argc--; argv++;
	    if (*argv == 0) {
	       (void) fprintf (stderr,
			       "udpsend: no add count given with '-n'.\n");
	       (void) exit (1);
	    }
	    nadd = atoi(*argv);
	 }
	 else {
	    nadd = atoi (ptr);
	    *ptr = 0;
	 }
	 break;

      case 'o':		/* output (info) file */
      case 'O':
	 oflag = YES;
	 if (*ptr == 0) {
	    argc--; argv++;
	    if (*argv == 0) {
	       (void) fprintf (stderr,
			       "udpsend: no output file given with '-o'.\n");
	       (void) exit (1);
	    }
	    (void) strcpy (outfile, *argv);
	 }
	 else {
	    (void) strcpy (outfile, ptr);
	    *ptr = 0;
	 }
	 break;
			
      case 'p':		/* packet size */
      case 'P':
	 if (*ptr == 0) {
	    argc--; argv++;
	    if (*argv == 0) {
	       (void) fprintf (stderr,
			       "udpsend: no packet size given with '-p'.\n");
	       (void) exit (1);
	    }
	    pktsize = atoi (*argv);
	 }
	 else {
	    pktsize = atoi (*argv);
	    *ptr = 0;
	 }
	 break;

      case 's':		/* sleep time between packets (ms) */
      case 'S':
	 sleepflag = YES;
	 if (*ptr == 0) {
	    argc--; argv++;
	    if (*argv == 0) {
	       (void) fprintf (stderr,
			       "udpsend: no sleep time given with '-s'.\n");
	       (void) exit (1);
	    }
	    sleeptime = atoi(*argv);
	 }
	 else {
	    sleeptime = atoi (ptr);
	    *ptr = 0;
	 }
	 break;

      case 't':		/* number of trains */
      case 'T':
	 if (*ptr == 0) {
	    argc--; argv++;
	    if (*argv == 0) {
	       (void) fprintf (stderr,
			       "udpsend: number of trains not given with '-t'.\n");
	       (void) exit (1);
	    }
	    ntrain = atoi(*argv);
	 }
	 else {
	    ntrain = atoi (ptr);
	    *ptr = 0;
	 }
	 break;

      case 'v':		/* verbose */
      case 'V':
	 verbose = YES;
	 break;

      default:
	 (void) fprintf (stderr,
			 "udpsend: Unknown option '%c'\n", ptr[-1]);
	 (void) exit (1);

      }
      argc--; argv++;
   }

   /* if gateway option given, find out the IP address of gateway */
   if (gway) {
      rr.rr_lsrr_gwy.s_addr = gethostaddr(gateway);
      if (rr.rr_lsrr_gwy.s_addr == 0) {
	 (void) fprintf (stderr,
			 "udpsend: %s: unknown gateway\n", gateway);
	 (void) exit (1);
      }
   }

   /* if no destination host given,
    * let the local host be the final destination */
   if (!dsthost)
      if (gethostname (hostname, MAXHNAME) < 0) {
	 perror ("udpsend: gethostname");
	 (void) exit (1);
      }

   /* set the destination as far as sender is concerned
    * i.e. echo host or the final destination host */
   rr.rr_lsrr_dst.s_addr = gethostaddr(hostname);
   if (rr.rr_lsrr_dst.s_addr == 0) {
      (void) fprintf (stderr, "udpsend: %s: unknown host\n",
		      hostname);
      (void) exit (1);
   }

   /* If no output file given, ask for it */
   if (!oflag) {
      (void) fprintf (stdout, "Output File? ");
      (void) fflush (stdout);
      (void) gets(outfile);
   }

   /* filename '-' indicates standard output */
   if (!(strcmp (outfile, "-"))) {
      fd = stdout;
      (void) strcpy (outfile, "Standard Output");
      goto nooutfile;
   }

   if (stat (outfile, &statb) < 0)
      /* If the file does not exist */
      fd = fopen (outfile, "w");
   else
      /* File exists; may be also check for directory */
      fd = fopen (outfile, "a");

   /* at this time, may be we should ask the user if he would like
    * to rewrite the file instead of appending, or may be even give
    * another file name */

   if (fd == NULL) {
      (void) fprintf (stderr, "can't open file %s\n", outfile);
      perror ("udpsend: fopen");
      (void) exit (1);
   }

nooutfile:

   /* create a socket to send packets */
   so = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP);
   if (so < 0) {
      perror ("udpsend: socket");
      (void) exit (1);
   }

   /* if packets to be routed via gateway, set up IP_LSRR option */
   if (gway)
      if (setsockopt (so, IPPROTO_IP, IP_OPTIONS, (char *)&rr,
		      sizeof(rr)) < 0) {
	 perror ("udpsend: IP_options");
	 (void) exit (1);
      }

   if (verbose) (void) printf ("Past setting options\n");

   /* initialize address/port number */
   bzero ((char *) &dst, sizeof (dst));
   dst.sin_family = AF_INET;
   dst.sin_addr = rr.rr_lsrr_dst;
   if (midhost)
      dst.sin_port = htons ((u_short) MIDPORT);
   else
      dst.sin_port = htons ((u_short) RECVPORT);

   /* assign the above address/port numbers permanently with socket */
   if (connect (so, (struct sockaddr *) &dst, sizeof (dst)) < 0) {
      perror ("udpsend: connect");
      (void) exit (1);
   }

   if (verbose) (void) printf ("Past connect\n");

   /* Write lots of information about the conenction */
   (void) fprintf (fd, "Experiment file name: %s\n", outfile);

   if (gway)
      (void) fprintf (fd, "Gateway Name: %s\n", gateway);
   else
      (void) fprintf (fd, "No gateway used\n");

   if (dsthost)
      (void) fprintf (fd, "Destination host: %s\n", hostname);
   else
      (void) fprintf (fd, "Destination on the local host\n");

   if (midhost)
      (void) fprintf (fd, "It is a 2-way connection\n");

   (void) fprintf (fd, "Number of trains = %d\n", ntrain);
   (void) fprintf (fd, "Number of packets = %d\n", npkt);
   (void) fprintf (fd, "Bulk-size = %d\n", nbulk);

   (void) fprintf (fd, "Sleep between packets : %u\n", sleeptime);
   (void) fprintf (fd, "Number of dummy additions : %d\n", nadd);
   (void) fprintf (fd, "Inter-train gap (sec) : %d\n", gap);

   (void) fprintf (fd, "Packet size: %d\n", pktsize);

   (void) gettimeofday(tp, (struct timezone *)0);
   (void) fprintf (fd, "experiment starting time = %s\n",
		   ctime((time_t *) &tp->tv_sec));
   (void) fprintf (fd, "Starting experiment.....\n\n");
   (void) fflush (fd);

   dcmax = (npkt+19)/20;
   /* 
    * Arbitrarily decided that in the verbose mode we will write out
    * a message after approx. every 5% of the packets have been sent.
    */

   sleeptime *= 1000;
   /*
    * I expect user to give sleep time in milliseconds. But the library
    * routine expects the number to be in micro-seconds. So I multiply
    * by 1000 here.
    */

   /* initial part of data is actually the log record */
   udprecord = (struct udprecord *) data;
   bzero((char *) udprecord, sizeof (struct udprecord));

   for (nt = 1; nt <= ntrain; nt++) {
      for (nc = 1; nc <= npkt; nc++) {
	 static int cc;

	 if (verbose)
	    if (++dcnt == dcmax) {
	       dcnt = 0;
	       (void) printf ("udpsend: in while loop %d\n", nc);
	    }

	/* Shifting the timestamp to inside the bulk sending -- SB. */
	/* it was originally here -- SB. */
	 /* send a bulk */
	 for (nb = 1; nb <= nbulk; nb++) {

	/* it has been shifted to here -- SB. */
	 /* get local timestamp and put it in the log record */
	 (void) gettimeofday (tp, (struct timezone *) 0);
	 udprecord->snd.tv_sec = htonl ((u_long) tp->tv_sec);
	 udprecord->snd.tv_usec = htonl ((u_long) tp->tv_usec);

	    udprecord->num = htonl ((u_long) pktsend+1);
	    cc = write (so, data, pktsize);
	    if (cc < 0) {
	       (void) fprintf (stderr,
			       "Packet number %d\n", nc);
	       perror("udpsend: write");
	       write_error = YES;
	       break;
	    }
	    pktsend++;		/* number of packets sent */
	 }
	 if (write_error) break;

	 /* do dummy additions */
	 if (nadd) {int i; for (i=0; i<nadd; i++);}

	 /* sleep between two successive bulks */
	 if (sleepflag) (void) usleep (sleeptime);
      }
      dcnt = 0;

      if (write_error) break;

      /* gap between two successive packet trains */
      if (gap) (void) sleep ((unsigned) gap);
   }

   (void) fprintf (fd, "Number of packets sent: %d\n", pktsend);
   (void) gettimeofday (tp, (struct timezone *) 0);
   (void) fprintf (fd, "experiment closing time = %s\n",
		   ctime((time_t *) &tp->tv_sec));
   (void) fclose (fd);

   /*
    * Send 5 packets with packet number to be 0. The receiver would
    * stop receiving packets and exit after it sees the first
    * packet with packet number 0. Other packets would be just
    * ignored.
    */
   {
      int	i;
      for (i = 1; i <= 5; i++) {
	 int	cc;
	 udprecord->num = 0;
	 cc = write (so, data, pktsize);
	 if (cc < 0) break;
	 if (sleeptime < 10000) sleeptime = 10000;
	 (void) usleep (sleeptime);
      }
   }
   (void) close (so);
}


usage()
{
   (void) fprintf (stderr, "usage: udpsend [-hv] <more options>\n");
   (void) fprintf (stderr, "The options are: \n");
   (void) fprintf (stderr, "\t -b \t bulk-size (number of packets sent together)\n");
   (void) fprintf (stderr, "\t -c \t packet count (actually number of times a bulk is sent)\n");
   (void) fprintf (stderr, "\t -d|m \t destination host\n");
   (void) fprintf (stderr, "\t -g \t gateway\n");
   (void) fprintf (stderr, "\t -i \t inter-train gap in seconds\n");
   (void) fprintf (stderr, "\t -n \t number of dummy additions\n");
   (void) fprintf (stderr, "\t -o \t output file\n");
   (void) fprintf (stderr, "\t -p \t packet size\n");
   (void) fprintf (stderr, "\t -s \t sleep between packets (in msec.)\n");
   (void) fprintf (stderr, "\t -t \t number of packet trains\n");

   (void) fprintf (stderr, "Single letter options are: \n");
   (void) fprintf (stderr, "\t -h \tHelp: produces this output\n");
   (void) fprintf (stderr, "\t -v \tVerbose; give info about progress\n");
   (void) exit (1);
}

/*
 * Return the address (in network byte order) of the requested host.
 *
 * We return 0 if we can't find anything.
 *
 * Arguably, this should return a struct in_addr, but 4.3 doesn't
 * seem to be able to make up it's mind.
 */

in_addr_t
gethostaddr(name)
     char *name;
{
   in_addr_t addr;
   register struct	hostent *hp;

   if ((addr = (in_addr_t)inet_addr (name)) != -1)
      return (addr);

   hp = gethostbyname(name);
   if (hp != NULL)
      return (*(in_addr_t *)hp->h_addr);
   else
      return (0);
}



#ifdef MIPS

/*
 * Copyright (c) 1989 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that: (1) source distributions retain this entire copyright
 * notice and comment, and (2) distributions including binaries display
 * the following acknowledgement:  ``This product includes software
 * developed by the University of California, Berkeley and its contributors''
 * in the documentation or other materials provided with the distribution
 * and in all advertising materials mentioning features or use of this
 * software. Neither the name of the University nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)usleep.c	5.5 (Berkeley) 6/1/90";
#endif /* LIBC_SCCS and not lint */

#define	USPS	1000000		/* number of microseconds in a second */

#define	setvec(vec, a) \
vec.sv_handler = a; vec.sv_mask = vec.sv_onstack = 0

static int ringring;

usleep (useconds)
     unsigned int useconds;
{
   register struct itimerval *itp;
   struct itimerval itv;
   struct sigvec vec, ovec;
   int omask;
   static void sleephandler();

   itp = &itv;
   if (!useconds)
      return;
   timerclear(&itp->it_interval);
   timerclear(&itp->it_value);
   itp->it_value.tv_sec = useconds / USPS;
   itp->it_value.tv_usec = useconds % USPS;
   setvec(vec, sleephandler);
   (void) sigvec(SIGALRM, &vec, &ovec);
   omask = sigblock(sigmask(SIGALRM));
   ringring = 0;
   (void) setitimer(ITIMER_REAL, itp, (struct itimerval *)0);
   while (!ringring)
      (void) sigpause(omask &~ sigmask(SIGALRM));
   (void) sigvec(SIGALRM, &ovec, (struct sigvec *)0);
   (void) sigsetmask(omask);
}

static void
sleephandler()
{
   ringring = 1;
}

#endif MIPS

